# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

#DONE: Bad orthogonal drawing

bl_info = {"name": "Draw xray",
           "description": "Draw xray mesh",
           "author": "JoseConseco",
           "version": (3, 8),
           "blender": (3, 6, 0),
           "location": "3D View(s) -> Top Bar -> Viewport overlays",
           "warning": "",
           "doc_url": "https://youtu.be/srUthznXDi4",
           "tracker_url": "https://discord.gg/cxZDbqH",
           "category": "3D View"
           }

import bpy
import ctypes
import bmesh
import gpu
from bpy_extras import view3d_utils
from gpu_extras.batch import batch_for_shader
from bpy.app.handlers import persistent
from mathutils import Vector, Matrix
from mathutils.bvhtree import BVHTree
import math
import numpy as np

#DONE: fix the ortho - kind of done
#DONE: maybe apply shrink wrap as eval mesh-  verts co?
#WONTDO: in multi edited meshes
#DONE: snap to target/offset as bpy.types.Obj property - remake snappint for that
#DONE: fix no bmesh, when opening scene. Maybe disable xray on quit/load?
#DONE: fix blinking mirror mod


BATCH_FACES = None
EDGES_BATCH = None
VERTS_BATCH = None
OBJ_MW = None
OBJ_MW_NP = None
PAUSE_HANDLERS = False #to avoid recursion
LAST_ACTIVE_OBJ = None
CACHED_OPERATOR_ID = ''  # to avoid recursion
FORCE_UPDATE_XRAY = False # not need anymore?
TARGET_BVH_LIST = None
AREA_ID = None # to keep track of active area  - from which xray was called
INIT_EDIT_MODE = 'OBJECT'  # help to track first jump to edit mode (so we do not snap on edit mode enter)

DRAW_POINTS = [] # debu

SINGLE_SKIP_UPDATE = False # helper to avoid snapping in some corncer cases

pts_shader = gpu.shader.from_builtin('UNIFORM_COLOR')

if bpy.app.version < (3, 5, 0):

    vertex_shader = '''
        uniform mat4 viewProjectionMatrix;
        uniform mat4 objMatrixWorld;
        uniform float bias_z;
        uniform vec3 camPos;

        in vec3 pos;
        in vec4 col;
        in vec3 nrm;

        out vec4 outPos;
        out vec4 color;
        out float oDot;
        void main()
        {
            color = col;
            vec4 pos4 = objMatrixWorld * vec4(pos, 1.0f);
            vec4 nrm4 = objMatrixWorld * vec4(nrm, 0.0);
            oDot = dot(normalize(nrm4.xyz), normalize(camPos-pos4.xyz));
            outPos = viewProjectionMatrix * pos4;
            outPos.z = outPos.z - bias_z/outPos.z; // counter w division for shift
            gl_Position = outPos;
        }
    '''

    fragment_shader = '''
        in vec4 color;
        in float oDot;
        out vec4 fragColor;
        void main()
        {
            vec4 out_color;
            if (oDot>0.0)
            {
                out_color = vec4(color.xyz * (0.3 + 0.7*oDot), color.w);
            }
            else
            {
                out_color = vec4(color.xyz * (0.7*oDot), 0.0);
            }
            fragColor = out_color;
        }
    '''

    shader = gpu.types.GPUShader(vertex_shader, fragment_shader)

else: # use new gpu for blender 3.5 and above

    # Shader that will offset faces more toward cam
    #############################################

    vert_out = gpu.types.GPUStageInterfaceInfo("my_draw_xray")
    vert_out.smooth('VEC4', "outPos")
    vert_out.smooth('VEC4', "color")
    vert_out.smooth('FLOAT', "oDot")

    shader_info = gpu.types.GPUShaderCreateInfo()
    shader_info.typedef_source(
                "struct CData {\n"
                "  mat4 ObjMatrixWorld;\n"
                "};\n"
            )

    shader_info.push_constant('MAT4', "viewProjectionMatrix")
    # shader_info.push_constant('MAT4', "objMatrixWorld") # cant be uniform (push_constant) due to uniform 128 size limit....
    # shader_info.uniform_buf(0, 'MAT4', 'objMatrixWorld')
    shader_info.uniform_buf(0, "CData", "ubo_vars")
    shader_info.push_constant('FLOAT', "bias_z")
    shader_info.push_constant('VEC3', "camPos")


    shader_info.vertex_in(0, 'VEC3', "pos")
    shader_info.vertex_in(1, 'VEC4', "col")
    shader_info.vertex_in(2, 'VEC3', "nrm")
    shader_info.vertex_out(vert_out)
    shader_info.fragment_out(0, 'VEC4', "FragColor")

    shader_info.vertex_source("""
    // uniform mat4 viewProjectionMatrix;
    // uniform mat4 objMatrixWorld;
    // uniform float bias_z;
    // uniform vec3 camPos;

    // in vec3 pos;
    // in vec4 col;
    // in vec3 nrm;

    // vec4 outPos; // XXX: was not out...
    // out vec4 color;
    // out float oDot;

    void main()
    {
        color = col;
        vec4 pos4 = ubo_vars.ObjMatrixWorld * vec4(pos, 1.0f);
        vec4 nrm4 = ubo_vars.ObjMatrixWorld * vec4(nrm, 0.0);
        oDot = dot(normalize(nrm4.xyz), normalize(camPos-pos4.xyz));
        outPos = viewProjectionMatrix * pos4;
        outPos.z = outPos.z - bias_z/outPos.z; // counter w division for shift
        gl_Position = outPos;
    }
    """)

    shader_info.fragment_source("""
    // in vec4 color;
    // in float oDot;
    // out vec4 FragColor;
    void main()
    {
        vec4 out_color;
        if (oDot>0.0)
        {
            out_color = vec4(color.xyz * (0.3 + 0.7*oDot), color.w);
        }
        else
        {
            out_color = vec4(color.xyz * (0.7*oDot), 0.0);
        }
        FragColor = out_color;
    }
    """)


    shader = gpu.shader.create_from_info(shader_info)
    del vert_out
    del shader_info

# end if


# ************************  workaround for uniform 128 size limit in gpu  ****************************************************
# from /mesh_snap_utilities_line/snap_context_l/mesh_drawing.py#L252
class _UBO_struct(ctypes.Structure):
    _pack_ = 16
    _fields_ = [
        ("ObjMatrixWorld", 4 * (4 * ctypes.c_float)),
    ]
UBO_data = _UBO_struct()
UBO = gpu.types.GPUUniformBuf(gpu.types.Buffer("UBYTE", ctypes.sizeof(UBO_data), UBO_data))

# END workaround  ****************************************************


def draw_callback_xray(self, context):
    active_obj = bpy.context.active_object
    if active_obj is None or not bpy.context.space_data.overlay.show_overlays:
        return

    scn_xray_props = context.scene.xray_props
    obj_xray_props = active_obj.xray_props
    shared_props = scn_xray_props if obj_xray_props.settings_mode == 'GLOBAL' else obj_xray_props
    if shared_props.draw_overlay == False:
        return

    if active_obj.type == 'MESH' and active_obj.mode in scn_xray_props.show_in_mode and BATCH_FACES:
        theme = bpy.context.preferences.themes['Default']
        g_vertex_size = theme.view_3d.vertex_size

        # for blender 3.5 and up only
        gpu.state.blend_set("ALPHA") # on=ALPHA # bgl.glEnable(bgl.GL_BLEND)
        gpu.state.line_width_set(3.0) # shader.uniform_float("lineWidth", 3.0)
        gpu.state.point_size_set(g_vertex_size + 1) # bgl.glPointSize(g_vertex_size + 1)
        gpu.state.face_culling_set("BACK") # bgl.glCullFace(bgl.GL_BACK)
        is_perspective = bpy.context.region_data.is_perspective
        if is_perspective:
            gpu.state.depth_test_set("LESS") # bgl.glEnable(bgl.GL_DEPTH_TEST)
        else: #hack cos I duno how fix bias in ortho
            gpu.state.depth_test_set("NONE") # bgl.glDisable(bgl.GL_DEPTH_TEST)

        extra_face_offset = 0.0
        if is_perspective:
            camera_pos = bpy.context.region_data.view_matrix.inverted().translation
            offset = scn_xray_props.drawOffset/50
            extra_face_offset = offset / 10
        else:
            # cam_dir_z = Vector((view_mat[2][0], view_mat[2][1], view_mat[2][2]))
            qat = context.region_data.view_rotation
            cam_dir_z = qat @ Vector((0, 0, 1))  # rotate Z vector by quat
            camera_pos = cam_dir_z * 100
            offset = 0.0

        shader.bind()
        shader.uniform_float("viewProjectionMatrix", context.region_data.perspective_matrix)
        # shader.uniform_float("objMatrixWorld", OBJ_MW) # complains: not enough mem in 128 uniform limit. Consider using UBO

        shader.uniform_block("ubo_vars", UBO)

        # mw_buff = gpu.types.GPUUniformBuf(np.array(active_obj.matrix_world, dtype=np.float32).tobytes())
        # shader.uniform_block("objMatrixWorld", mw_buff) # complains: not enough mem in 128 uniform limit. Consider using UBO)
        shader.uniform_float("bias_z", offset if extra_face_offset == 0.0 else offset - extra_face_offset) # hack replaces GL_POLYGON_OFFSET_FILL
        shader.uniform_float("camPos", camera_pos)
        # with gpu.matrix.push_pop():
            # gpu.matrix.multiply_matrix(active_obj.matrix_world) # why it wont affect the shader?

        BATCH_FACES.draw(shader)

        shader.uniform_float("bias_z", offset) # case for faces and verts
        EDGES_BATCH.draw(shader)
        if bpy.context.tool_settings.mesh_select_mode[0] and active_obj.mode == 'EDIT':
            VERTS_BATCH.draw(shader)
        # restore opengl defaults


        gpu.state.depth_test_set("NONE") # bgl.glDisable(bgl.GL_DEPTH_TEST)
        gpu.state.face_culling_set("NONE") # XXX: bgl.glDisable(bgl.GL_CULL_FACE)
        gpu.state.blend_set("NONE") # bgl.glDisable(bgl.GL_BLEND)
        # bgl.glPolygonMode(bgl.GL_FRONT_AND_BACK, bgl.GL_FILL)
        gpu.state.line_width_set(1.0) # shader.uniform_float("lineWidth", 3.0)
        gpu.state.point_size_set(5) # bgl.glPointSize(1)

        if DRAW_POINTS: # XXX:  DEBUG
            # draw points using pts_shader
            batch = batch_for_shader(pts_shader, 'POINTS', {"pos": DRAW_POINTS})
            pts_shader.uniform_float("color", (1, 0.5, 0.5, 1.0))
            batch.draw(pts_shader)

        gpu.state.point_size_set(1) # bgl.glPointSize(1)



def get_collection_objects(coll, active_obj):
    col_objs = set()
    for obj in coll.all_objects: # obj from coll and child colls
        if obj.visible_get() and obj.display_type in ('SOLID','TEXTURED') and obj != active_obj: # new algo can handle everything (curves, groups, ..)
            col_objs.add(obj.name)
    # for coll in coll.children_recursive:
    #     for obj in coll.objects:
    #         if obj.type == 'MESH':
    #             col_objs.append(obj)
    return col_objs



def get_nested_obj_depsgrapth_bm(bm, depsgraph, src_objs : set, apply_modifiers=True, matrix=None):
    ''' handles everything : groups, curves etc
       Returns: None - modifies passed bm '''

    def is_object_instance_from_selected(object_instance, base_ob):
        # For instanced objects we check selection of their instancer (more accurately: check
        # selection status of the original object corresponding to the instancer).
        if object_instance.parent: #  it refers to group instance, rather than  rad array modifier.
            return object_instance.parent.original.name in base_ob
        # For non-instanced objects we check selection state of the original object.
        return object_instance.object.original.name in base_ob

    def get_obj_for_instance(object_instance, apply_modifiers):
        if apply_modifiers:
            return object_instance.object
        else:
            return object_instance.object.original

    '''Works in groups too in recursive way. apply_modifiers=False => modes get ignored competly'''
    for dep_obj_info in depsgraph.object_instances:
        if not is_object_instance_from_selected(dep_obj_info, src_objs):
            continue # We only export selected objects

        orig_obj = get_obj_for_instance(dep_obj_info, apply_modifiers)
        if not orig_obj.data or orig_obj.type != 'MESH' or orig_obj.display_type not in ('SOLID','TEXTURED'): # skip for non-geometry objects.
            continue

        # print('inter on: '+orig_obj.name)
        # bm.from_mesh(orig_obj.to_mesh())   # with modifiers
        # bm.transform(dep_obj_info.matrix_world)

        me = orig_obj.to_mesh()
        # print(f"{len(me.vertices)=}")
        me.transform(dep_obj_info.matrix_world)
        bm.from_mesh(me) # should join all meshes into one bm

        # me.to_mesh_clear()
        orig_obj.to_mesh_clear()

    if matrix:
        bm.transform(matrix)



def debug_mesh(bm):
    mesh = bpy.data.meshes.new('temp_mesh')
    bm.to_mesh(mesh)
    new_obj = bpy.data.objects.new('temp_obj', mesh)
    bpy.context.scene.collection.objects.link(new_obj)



def get_obj_mesh_bvht(context, depsgraph, applyModifiers=True, world_space=True):
    active_obj = context.active_object

    obj_xray_props = active_obj.xray_props
    scn_xray_props = context.scene.xray_props
    shared_xray_props = scn_xray_props if obj_xray_props.settings_mode == 'GLOBAL' else obj_xray_props

    if applyModifiers:
        if world_space:
            # #? wont work eg with shrink wrap mod, but fast....
            # obj.data.transform(obj.matrix_world)
            # depsgraph.update() #fixes bad transformation baing applied to obj
            # bvh = BVHTree.FromObject(obj, depsgraph)  #? not required to get with mod: obj.evaluated_get(depsgraph)
            # obj.data.transform(obj.matrix_world.inverted())
            # return bvh

            #* better but slower - even 5-10 times (0.05 sec), wont work on non meshes (curves?)
            # get merged version of scene objects with modifiers.   - skip non mesh objs
            # NOTE: this is slow, but works for all objects (even curves)

            active_obj_mw = active_obj.matrix_world  # bring it into active obj (retopo obj) space...

            if shared_xray_props.snap_target == 'OBJECT':
                snap_obj  = shared_xray_props.snap_object
                if not shared_xray_props.snap_object or not shared_xray_props.snap_object.visible_get() or shared_xray_props.snap_object == active_obj:
                    return None

                bm = bmesh.new()   # create an empty BMesh
                # handles groups, geo nodes generated objs, mods, curves? etc.
                get_nested_obj_depsgrapth_bm(bm, depsgraph, {snap_obj.name,}, apply_modifiers=True, matrix=active_obj_mw.inverted())
                bm.normal_update() # or else we get bad normals in snapping!
                bvh = BVHTree.FromBMesh(bm)  # ? not required to get with mod: obj.evaluated_get(depsgraph)
                bm.free()  # free and prevent further access

                # WARNING: below option would say 'empty' instance coll has no mesh... even if it has one inside inst_col
                # obj_eval = snap_obj.evaluated_get(depsgraph)
                # bm = bmesh.new()   # create an empty BMesh
                # bm.from_mesh(obj_eval.to_mesh())   # with modifiers
                # bm.transform(active_obj_mw.inverted() @ snap_obj.matrix_world)
                # bvh = BVHTree.FromBMesh(bm)  # ? not required to get with mod: obj.evaluated_get(depsgraph)
                # bm.free()  # free and prevent further access
                # obj_eval.to_mesh_clear()
                print('Updated BVHTree')
                return bvh

            else:
                # NOTE: bmesh can merge multiple src meshes!
                # bm = bmesh.new()  # create an empty BMesh - and join all meshes into int
                # bm.from_mesh( mesh1 ) # Add first mesh
                # bm.from_mesh( mesh2 ) # Add 2nd mesh
                snap_objs = []
                if shared_xray_props.snap_target == 'COLLECTION':
                    if shared_xray_props.snap_collection in context.scene.collection.children_recursive:
                        snap_objs = get_collection_objects(shared_xray_props.snap_collection, active_obj)
                else:
                    snap_objs = {o.name for o in context.scene.objects if o.visible_get() and o.display_type in ('SOLID','TEXTURED') and o != active_obj}
                if not snap_objs:
                    print('Xray: No snap objects found')
                    return None

                bm = bmesh.new()
                # for o in snap_objs: # WARNING: may be slow (O^2 :  for snap_objs: for all eval objs..)
                get_nested_obj_depsgrapth_bm(bm, depsgraph, snap_objs, apply_modifiers=True)
                bm.transform(active_obj_mw.inverted())
                bm.normal_update() # or else we get bad normals in snapping!
                bvh = BVHTree.FromBMesh(bm)  # ? not required to get with mod: obj.evaluated_get(depsgraph)
                # debug_mesh(bm)
                bm.free()  # free and prevent further access

                # NOTE: ok but empties instances would not be evaluated as groups, curves - give error 'Obj does not have geometry data'
                # bm = bmesh.new()
                # eval_objs = [] # for ev_obj.to_mesh_clear
                # for o in snap_objs:
                #     if o != active_obj and o.visible_get():
                #         obj_ev = o.evaluated_get(depsgraph)
                #         eval_objs.append(obj_ev)
                #         me = obj_ev.to_mesh()
                #         me.transform(o.matrix_world)
                #         bm.from_mesh(me) # should join all meshes into one bm
                # bm.transform(active_obj_mw.inverted())
                # bm.normal_update()
                # bvh = BVHTree.FromBMesh(bm)  # ? not required to get with mod: obj.evaluated_get(depsgraph)
                # bm.free()  # free and prevent further access
                # for obj_eval in eval_objs:
                #     obj_eval.to_mesh_clear()

                print('Updated BVHTree')
                return bvh
        else:
            return BVHTree.FromObject(snap_obj, depsgraph) #with modes
    else:
        if world_space:
            # 4 times slower than data.transform
            #bvh1 =  BVHTree.FromPolygons([obj.matrix_world @ v.co for v in obj.data.vertices], [p.vertices for p in obj.data.polygons])
            #bmesh - same time as data.transform
            snap_obj.data.transform(snap_obj.matrix_world)
            bvh = BVHTree.FromPolygons([v.co for v in snap_obj.data.vertices], [p.vertices for p in snap_obj.data.polygons])
            snap_obj.data.transform(snap_obj.matrix_world.inverted())
            print('Updated BVHTree')
            return bvh
        else:
            print('Updated BVHTree')
            return BVHTree.FromPolygons([v.co for v in snap_obj.data.vertices], [p.vertices for p in snap_obj.data.polygons])


class ScnDrawXrayProps(bpy.types.PropertyGroup):
    def DrawXrayUpdate(self,context):
        # if context.active_object and \
        #     (self.draw_overlay == False \
        #     or (self.settings_mode == 'GLOBAL' and not context.scene.xray_props.enable_snapping)
        #     or (self.settings_mode == 'PER_OBJ' and not context.active_object.xray_props.enable_snapping)):
        global PAUSE_HANDLERS
        PAUSE_HANDLERS = False
        handle_handlers_draw_xray()
        refresh_draw_buffers()
        global AREA_ID
        AREA_ID = context.area.as_pointer() if self.enable_xray else None

        if context.active_object and context.active_object.type == 'MESH':
            context.active_object.data.update_tag()

    def refresh_buff(self,context):
        refresh_draw_buffers()
        # context.active_object.data.update_tag()

    def clear_snap_cache(self,context):
        global TARGET_BVH_LIST
        TARGET_BVH_LIST = None

    def toggle_snap(self,context):
        if not self.enable_snapping:
            self.clear_snap_cache(context)
        global AREA_ID, SINGLE_SKIP_UPDATE
        AREA_ID = context.area.as_pointer() if self.enable_xray and self.enable_snapping else None
        SINGLE_SKIP_UPDATE = True
        # context.active_object.data.update_tag()

    def skip_snap(self,context):
        global SINGLE_SKIP_UPDATE
        SINGLE_SKIP_UPDATE = True

    def verify_snap_target(self,context):
        if self.snap_target == 'OBJECT':
            self['snap_collection'] = None
        elif self.snap_target == 'COLLECTION':
            self['snap_object'] = None
        else:
            self['snap_object'] = None
            self['snap_collection'] = None

        global SINGLE_SKIP_UPDATE
        SINGLE_SKIP_UPDATE = True
        self.clear_snap_cache(context)
        # context.active_object.data.update_tag()

    enable_xray: bpy.props.BoolProperty(name='Enable Xray', description='Enable Xray', default=False, update=DrawXrayUpdate)

    show_in_mode: bpy.props.EnumProperty(name='Show in Mode', description='Show in Mode',
                                         items=[('EDIT', 'Edit', 'Edit Mode'),
                                                ('SCULPT', 'Sculpt', 'Sculpt Mode'),
                                                ('PAINT', 'Paint', 'Paint Mode'),
                                                ('OBJECT', 'Object', 'Object Mode')]
                                                , options={'ENUM_FLAG'}, default={'EDIT'})


    fold_overlay: bpy.props.BoolProperty(name='Fold Settings', description='Fold Settings', default=False)
    draw_overlay: bpy.props.BoolProperty(name='Draw Xray Overlay', description='Draw Overlay On top of lowpoly object.\nComes with optional snapping feature (onliy in paid ver)', update=refresh_buff, default=True)

    drawOffset: bpy.props.FloatProperty(name="Depth Bias", description="Moves rendering of mesh closer to camera (does not affect mesh geometry)", min=0.0001, soft_max=1.0, default=0.2, subtype='FACTOR', update=refresh_buff)
    polygon_opacity: bpy.props.FloatProperty(name="Face opacity", description="Face opacity", min=0.0, max=1.0, default=0.5, subtype='PERCENTAGE', update=refresh_buff)
    edgeOpacity: bpy.props.FloatProperty(name="Edge opacity", description="Edge opacity", min=0.0, max=1.0, default=0.5, subtype='PERCENTAGE', update=refresh_buff)
    face_color:  bpy.props.FloatVectorProperty(name="Face Color", subtype='COLOR', default=(0.1, 0.8, 0.0), min=0.0, max=1.0, description="color picker", update=refresh_buff)
    highlight_color:  bpy.props.FloatVectorProperty(name="Highlight Color", subtype='COLOR', default=(1.0, 0.8, 0.0), min=0.0, max=1.0, description="color picker", update=refresh_buff)

    # clear_cache: bpy.props.BoolProperty(name='Reload Snapping Cache', description='Reloads snapping cache - use if snapping does not work correctly, e.g. after one of snap target objects was hidden...', default=False, update=clear_snap_cache)
    fold_snap: bpy.props.BoolProperty(name='Fold Snap', description='Fold Snap', default=False)
    draw_modifiers: bpy.props.BoolProperty(name="Draw Modifiers", description="Draws retopo mesh with modifiers", default=False, update=refresh_buff)
    enable_snapping: bpy.props.BoolProperty(name="Snapping", description="Enable global snapping, for all objects on scene.\nEnabling this feature may result in a brief delay as the cache is built.", default=False, update=toggle_snap)
    snap_method: bpy.props.EnumProperty(name="Snap Method", description="Snap Method",
                                        items=[('PROJECT', 'Method: Surface Projected', ''),
                                               ('NEAREST', 'Method: Surface Nearest', ''),
                                               ], default='NEAREST', update=skip_snap) # , update=lambda self, context: context.active_object.data.update_tag()
    affect_neighbours: bpy.props.BoolProperty(name="Affect Neighbours", description="Toggle snapping of adjacent (non selected) vertices", default=False, update=skip_snap) # update=lambda self, context: context.active_object.data.update_tag())

    snap_offset: bpy.props.FloatProperty( name="Snap offset", description="Offset retopo mesh vertices above high-poly mesh surface.", default=0.01, soft_min=0.0, soft_max=0.1, update=lambda self, context: context.active_object.data.update_tag())
    # use_simple_snap: bpy.props.BoolProperty(name="Legacy Snap", description="Use simple snap method (faster, but less accurate)", default=False, update=lambda self, context: context.active_object.data.update_tag())

    snap_target: bpy.props.EnumProperty(name="Snap Method", description="Snap Method",
                                        items=[('OBJECT', 'Snap to: Object', 'Object'),
                                               ('COLLECTION', 'Snap to: Collection', 'Collection'),
                                               ('ALL', 'Snap to: All Objects', 'Snap To All objects except active one')], default='ALL', update=verify_snap_target)
    snap_object: bpy.props.PointerProperty(name='Snap Object', description="Default snap target object for all objects.", type=bpy.types.Object,
                                           update=verify_snap_target)
                                        # XXX: update=verify_snap_target, poll=lambda self, obj: obj.type == 'MESH' and obj != bpy.context.active_object)
    snap_collection: bpy.props.PointerProperty(name='Snap Collection', description="Default snap target collection for all objects.", type=bpy.types.Collection,
                                               update=verify_snap_target, poll=lambda self, col: col in bpy.context.scene.collection.children_recursive)

    snap_event: bpy.props.EnumProperty(name='Snap Event', description='Snap Mode',
                                       items=[('ALL', 'Event: Geometry & Selection Update', 'Always snap geometry (including geo select events)'),
                                              ('GEO_UPDATE', 'Event: Geometry Update', 'Snap only on geometry update')], default='GEO_UPDATE', update=skip_snap)

    # NOTE: obsolete (only cuold be used for sculpt mode)?
    # wrap_method: bpy.props.EnumProperty(name="Mode", description="Shrink wrap Mode", default="NEAREST_SURFACEPOINT",
    #                                           items=[('NEAREST_SURFACEPOINT', 'Nearest Surface point', ''),
    #                                                  ('PROJECT', 'Project', ''),
    #                                                  ('NEAREST_VERTEX', 'Nearest Vertex', ''),
    #                                                  ('TARGET_PROJECT', 'Target Normal Project', '')
    #                                                  ], update=verify_snap_target)
    # wrap_mode: bpy.props.EnumProperty(name='Wrap Mode', description='Wrap Mode',
    #     items=[('ON_SURFACE', 'On Surface', 'On Surface'),
    #         ('INSIDE', 'Inside', 'Inside'),
    #         ('OUTSIDE', 'Outside', 'Outside'),
    #         ('OUTSIDE_SURFACE', 'Outside Surface', 'Outside Surface'),
    #         ('ABOVE_SURFACE', 'Above surface', 'Above_surface')],
    #     default='OUTSIDE_SURFACE', update=verify_snap_target)
    #


class ObjDrawXrayProps(bpy.types.PropertyGroup):
    def refresh_buff(self, context):
        refresh_draw_buffers()
        context.active_object.data.update_tag()

    def clear_snap_cache(self,context):
        global TARGET_BVH_LIST
        TARGET_BVH_LIST = None

    def toggle_snap(self,context):
        if not self.enable_snapping:
            self.clear_snap_cache(context)
        global AREA_ID, SINGLE_SKIP_UPDATE
        AREA_ID = context.area.as_pointer() if self.enable_snapping else None
        SINGLE_SKIP_UPDATE = True
        # context.active_object.data.update_tag()

    def skip_snap(self,context):
        global SINGLE_SKIP_UPDATE
        SINGLE_SKIP_UPDATE = True

    def verify_snap_target(self,context):
        if self.snap_target == 'OBJECT':
            self['snap_collection'] = None
        elif self.snap_target == 'COLLECTION':
            self['snap_object'] = None
        else:
            self['snap_object'] = None
            self['snap_collection'] = None


        global SINGLE_SKIP_UPDATE
        SINGLE_SKIP_UPDATE = True
        self.clear_snap_cache(context)
        # context.active_object.data.update_tag()

    settings_mode: bpy.props.EnumProperty(name='Settings Mode', description='Settings Mode',
                                          items=[('GLOBAL', 'Use Global Settings', 'Global Settings'),
                                                 ('PER_OBJ', 'Unique (per object)', 'Unique Snap Settings for each unique Object')],
                                          default='GLOBAL')

    draw_overlay: bpy.props.BoolProperty(name='Draw Xray Overlay', default=True, update=refresh_buff)
    draw_modifiers: bpy.props.BoolProperty(name="Draw Modifiers", description="Draws retopo mesh with modifiers", default=False, update=refresh_buff)
    enable_snapping: bpy.props.BoolProperty(name="Snapping", description="Enable global snapping, for all objects on scene.\nEnabling this feature may result in a brief delay as the cache is built.", default=False, update=toggle_snap)
    snap_method: bpy.props.EnumProperty(name="Snap Method", description="Snap Method",
                                        items=[('PROJECT', 'Method: Surface Projected', ''),
                                               ('NEAREST', 'Method: Surface Nearest', ''),
                                               ], default='NEAREST', update=skip_snap) # , update=lambda self, context: context.active_object.data.update_tag()
    affect_neighbours: bpy.props.BoolProperty(name="Affect Neighbours", description="Toggle snapping of adjacent (non selected) vertices", default=False, update=skip_snap)  # , update=lambda self context: context.active_object.data.update_tag())
    snap_offset: bpy.props.FloatProperty( name="Snap offset", description="Offset retopo mesh vertices above high-poly mesh surface", default=0.01, soft_min=0.0, soft_max=0.1, update=lambda self, context: context.active_object.data.update_tag())

    snap_target: bpy.props.EnumProperty(name="Snap Method", description="Snap Method",
                                        items=[('OBJECT', 'Snap to: Object', 'Object'),
                                               ('COLLECTION', 'Snap to: Collection', 'Collection'),
                                               ('ALL', 'Snap to: All Objects', 'Snap To All objects except active one')], default='ALL', update=verify_snap_target)
    snap_object: bpy.props.PointerProperty(name='Snap Object', description="Default snap target object for all objects.", type=bpy.types.Object,
                                           update=verify_snap_target, poll=lambda self, obj: obj.type == 'MESH' and obj != bpy.context.active_object)
    snap_collection: bpy.props.PointerProperty(name='Snap Collection', description="Default snap target collection for all objects.", type=bpy.types.Collection,
                                               update=verify_snap_target, poll=lambda self, col: col in bpy.context.scene.collection.children_recursive)

    # NOTE: obsolete (only cuold be used for sculpt mode)?
    # wrap_method: bpy.props.EnumProperty(name="Mode", description="Shrink wrap Mode", default="NEAREST_SURFACEPOINT",
    #                                     items=[('NEAREST_SURFACEPOINT', 'Nearest Surface point', ''),
    #                                            ('PROJECT', 'Project', ''),
    #                                            ('NEAREST_VERTEX', 'Nearest Vertex', ''),
    #                                            ('TARGET_PROJECT', 'Target Normal Project', '')
    #                                            ])
    # wrap_mode: bpy.props.EnumProperty(name='Wrap Mode', description='Wrap Mode',
    #                                   items=[('ON_SURFACE', 'On Surface', 'On Surface'),
    #                                          ('INSIDE', 'Inside', 'Inside'),
    #                                          ('OUTSIDE', 'Outside', 'Outside'),
    #                                          ('OUTSIDE_SURFACE', 'Outside Surface', 'Outside Surface'),
    #                                          ('ABOVE_SURFACE', 'Above surface', 'Above_surface')],
    #                                   default='ABOVE_SURFACE')



def ShadingXrayPanel(self, context):
    if context.active_object and context.active_object.type == 'MESH':
        scn_xray_props = context.scene.xray_props
        obj_xray_props = context.active_object.xray_props
        shared_xray_props = scn_xray_props if obj_xray_props.settings_mode == 'GLOBAL' else obj_xray_props

        box = self.layout.box()
        row = box.row(align=True)
        row.prop(scn_xray_props, "enable_xray", expand=True, icon='MOD_SOLIDIFY', toggle=True)
        if not scn_xray_props.enable_xray:
            return

        row = box.row(align=True)
        row.prop(obj_xray_props, "settings_mode", expand=True)

        sub_box = box.box()
        sub_box.enabled = True # False for version without snap
        row = sub_box.row(align=True)
        row.prop(shared_xray_props, "draw_overlay", icon='CHECKBOX_HLT' if shared_xray_props.draw_overlay else 'CHECKBOX_DEHLT', emboss=False)
        row.prop(shared_xray_props, "draw_modifiers", icon="MODIFIER" if shared_xray_props.draw_modifiers else 'MODIFIER_OFF', icon_only=True, emboss=False)
        row.prop(scn_xray_props, "fold_overlay", icon='TRIA_DOWN' if not scn_xray_props.fold_overlay else 'TRIA_LEFT', text='', emboss=False, toggle=True)

        #* Overlay Drawing  Settings
        if shared_xray_props.draw_overlay and not scn_xray_props.fold_overlay:
            sub_col = sub_box.column()

            row = sub_col.row(align=True)
            row.prop(scn_xray_props, "show_in_mode", expand=False)
            row = sub_col.row(align=True)
            row.prop(scn_xray_props, "drawOffset")
            row = sub_col.row(align=True)
            row.prop(scn_xray_props, "polygon_opacity")
            row.prop(scn_xray_props, "edgeOpacity")
            row = sub_col.row(align=True)
            row.label(text="Faces color")
            row.prop(scn_xray_props, "face_color", text='')
            row = sub_col.row(align=True)
            row.label(text="Highlight color")
            row.prop(scn_xray_props, "highlight_color", text='')

        # Snapping
        sub_box = box.box()
        sub_box.enabled = True # False for version without snap
        # sub_box.label(text='Only in paid version')
        row = sub_box.row(align=True)
        row.prop(shared_xray_props, "enable_snapping", icon='CHECKBOX_HLT' if shared_xray_props.enable_snapping else 'CHECKBOX_DEHLT', emboss=False)
        row.prop(scn_xray_props, "fold_snap", icon='TRIA_DOWN' if not scn_xray_props.fold_snap else 'TRIA_LEFT', text='', emboss=False, toggle=True)
        if not scn_xray_props.fold_snap:
            col = sub_box.column()
            col.active = shared_xray_props.enable_snapping
            row = col.row(align=True)
            row.prop(shared_xray_props, "snap_method", text='')
            if shared_xray_props.snap_method == 'NEAREST':
                row.prop(shared_xray_props, "affect_neighbours", text='', icon='AUTOMERGE_ON' if shared_xray_props.affect_neighbours else 'AUTOMERGE_OFF')
            # row.prop(scn_xray_props, "use_simple_snap", icon='UNLOCKED' if scn_xray_props.snap_event == 'ALL' else 'LOCKED', text='')
            row = col.row(align=True)
            row.prop(shared_xray_props, "snap_target", text='')
            if shared_xray_props.snap_target == 'OBJECT':
                row.prop_search(shared_xray_props, "snap_object", context.scene, "objects", text='')
                picker_op = row.operator(XRAY_OT_ObjectPicker.bl_idname, icon='EYEDROPPER', text='')
                picker_op.use_scn_prop = obj_xray_props.settings_mode == 'GLOBAL'
            elif shared_xray_props.snap_target == 'COLLECTION':
                row.prop_search(shared_xray_props, "snap_collection", bpy.data, "collections", text='')
            # row.prop(scn_xray_props, "clear_cache", icon='FILE_REFRESH', text='')
            col.prop(shared_xray_props, "snap_offset")
            row = col.row(align=True)
            row.prop(scn_xray_props, "snap_event", icon='UNLOCKED' if scn_xray_props.snap_event == 'ALL' else 'LOCKED', text='')




class XRAY_OT_ObjectPicker(bpy.types.Operator):
    bl_idname = "xray.object_picker"
    bl_label = "Pick Object"
    bl_description = 'Pick Object under the cursor'
    bl_options = {'REGISTER', 'UNDO', 'INTERNAL'}

    use_scn_prop: bpy.props.BoolProperty(name='Scene prop', description='Pick scn.snap_object prop or obj.snap_object', default=True)

    @classmethod
    def poll(cls, context):
        return True

    def invoke(self, context, event):
        self.lmb_clicked = True
        self.obj = None
        context.window.cursor_modal_set('EYEDROPPER')
        context.window_manager.modal_handler_add(self)
        self.depsgraph = context.evaluated_depsgraph_get()
        return {"RUNNING_MODAL"}

    def scn_ray_cast(self, context, event):
        ''' using scn.ray_cast '''
        region = context.region
        rv3d = context.region_data
        coord = event.mouse_region_x, event.mouse_region_y
        view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, coord)
        ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord)
        # hit, loc, norm, idx, obj, mat = context.scene.ray_cast(context.view_layer, ray_origin, view_vector, distance=1.0e+6) #old
        hit, loc, norm, idx, obj, mat = context.scene.ray_cast(self.depsgraph, ray_origin, view_vector, distance=1.0e+6)
        return obj

    def modal(self, context, event):
        if event.type == 'MOUSEMOVE':
            self.obj = self.scn_ray_cast(context, event)
            context.area.header_text_set(self.obj.name if self.obj else None)

        elif event.type == "MIDDLEMOUSE":
            return {'PASS_THROUGH'}

        elif event.type == "LEFTMOUSE":
            if self.obj and self.obj.type == 'MESH':
                if self.use_scn_prop:
                    context.scene.xray_props.snap_object = self.obj
                else:
                    context.active_object.xray_props.snap_object = self.obj
                self.report({'INFO'}, f'Picked {self.obj.name}')
            else:
                self.report({'WARNING'}, 'Pick mesh type of object')
            context.window.cursor_modal_restore()
            context.area.header_text_set(None)
            return {"FINISHED"}

        elif event.type in {"RIGHTMOUSE", "ESC"}:
            context.window.cursor_modal_restore()
            context.area.header_text_set(None)
            return {"CANCELLED"}

        return {"RUNNING_MODAL"}


handle_SpaceView3D = None

@persistent
def xray_scene_update(scene, depsgraph):
    check_obj_updated(depsgraph)


@persistent
def DrawXrayPost(scn):
    handle_handlers_draw_xray()

@persistent
def DrawXrayLoadDisable(scn):
    #print('Disabling xray on load handerl')
    bpy.context.scene.xray_props['enable_snapping'] = False  #
    bpy.context.scene.xray_props.enable_xray = False  # to avoid bm edit error


def handle_handlers_draw_xray():
    global handle_SpaceView3D
    scn_xray_props = bpy.context.scene.xray_props
    if scn_xray_props.enable_xray:
        if handle_SpaceView3D is None:
            args = (ScnDrawXrayProps, bpy.context)  # u can pass arbitrary class as first param  Instead of (self, context)
            handle_SpaceView3D = bpy.types.SpaceView3D.draw_handler_add(draw_callback_xray, args, 'WINDOW', 'POST_VIEW')

        if xray_scene_update not in bpy.app.handlers.depsgraph_update_post:
            bpy.app.handlers.depsgraph_update_post.append(xray_scene_update)
    else:
        if handle_SpaceView3D is not None:
            bpy.types.SpaceView3D.draw_handler_remove(handle_SpaceView3D, 'WINDOW')
            handle_SpaceView3D = None

        if xray_scene_update in bpy.app.handlers.depsgraph_update_post:
            bpy.app.handlers.depsgraph_update_post.remove(xray_scene_update)



VERTICES = None
INDICES = None
NORMALS = None

def refresh_draw_buffers(update_pos_cache=True):
    global PAUSE_HANDLERS
    PAUSE_HANDLERS = True # just make sure this drawing wont fire geo updates in depsgraph.updates
    active_obj = bpy.context.active_object
    if not active_obj or active_obj.type != 'MESH':
        PAUSE_HANDLERS = False
        return
    scn_xray_props = bpy.context.scene.xray_props
    obj_xray_props = active_obj.xray_props
    shared_props = scn_xray_props if obj_xray_props.settings_mode == 'GLOBAL' else obj_xray_props
    if not shared_props.draw_overlay:
        PAUSE_HANDLERS = False
        return
    active_obj.update_from_editmode() # get latest selected verts  (otherwise it will use old sel)

    # if True:  # Polyquilt  fix but wont work - selection not updated if no modifiers, and  draw_modifiers == True
    if  shared_props.draw_modifiers: # includes shapekesy
        depsgraph = bpy.context.evaluated_depsgraph_get() # WARNING: somehow using DEPSGRAPH here, causes blink of modifiers
        # depsgraph.update()  # to get disabled mods/ update selection highlight if draw mods_on,
        obj_eval = active_obj.evaluated_get(depsgraph)
        mesh = obj_eval.to_mesh()
    else:
        mesh = active_obj.data

    theme = bpy.context.preferences.themes['Default']

    g_vertex_color = theme.view_3d.vertex
    # g_vertex_size = theme.view_3d.vertex_size
    g_wire_edit_color = theme.view_3d.wire_edit

    g_face_color = scn_xray_props.face_color
    highlight_color = scn_xray_props.highlight_color
    #print('Runinig buffer update')
    mesh.calc_loop_triangles()
    vert_count = len(mesh.vertices)
    global VERTICES, INDICES, NORMALS
    if update_pos_cache:
        VERTICES = np.empty((vert_count, 3), 'f')
        INDICES = np.empty((len(mesh.loop_triangles), 3), 'i')
        NORMALS = np.empty((vert_count, 3), 'f')
        # mesh.transform(active_obj.matrix_world) # XXX: since 3.3 (and 3.2?) this causes len(mesh.vertices) to be 0 :/
        # mesh.update()  #? recalc normals - brokens see above
        mesh.vertices.foreach_get("co", np.reshape(VERTICES, vert_count * 3))
        mesh.vertices.foreach_get("normal", np.reshape(NORMALS, vert_count * 3))
        mesh.loop_triangles.foreach_get("vertices", np.reshape(INDICES, len(mesh.loop_triangles) * 3))


    face_col = [(g_face_color[0], g_face_color[1], g_face_color[2], scn_xray_props.polygon_opacity/2) for _ in range(vert_count)]
    edge_col = [(g_wire_edit_color.r, g_wire_edit_color.g, g_wire_edit_color.b, scn_xray_props.edgeOpacity*0.7) for _ in range(vert_count)]
    vert_opacity = min(scn_xray_props.edgeOpacity+0.2, 1.0)
    vert_col = [(g_vertex_color.r, g_vertex_color.g, g_vertex_color.b, vert_opacity) for _ in range(vert_count)]

    if active_obj.mode == 'EDIT':
        for i,vert in enumerate(mesh.vertices):
            if vert.select:
                edge_col[i] = (highlight_color[0], highlight_color[1], highlight_color[2], scn_xray_props.edgeOpacity*0.7)
                vert_col[i] = (highlight_color[0], highlight_color[1], highlight_color[2], vert_opacity)
                face_col[i] = (highlight_color[0], highlight_color[1], highlight_color[2], scn_xray_props.polygon_opacity/2)

    global BATCH_FACES, EDGES_BATCH, VERTS_BATCH


    ob_mat = active_obj.matrix_world.transposed() # why trans?

    UBO_data.ObjMatrixWorld[0] = ob_mat[0][:]
    UBO_data.ObjMatrixWorld[1] = ob_mat[1][:]
    UBO_data.ObjMatrixWorld[2] = ob_mat[2][:]
    UBO_data.ObjMatrixWorld[3] = ob_mat[3][:]
    UBO.update(gpu.types.Buffer( "UBYTE", ctypes.sizeof(UBO_data), UBO_data))


    # mw_array = np.array(active_obj.matrix_world, 'f').transpose()  # into 4x4 matrix
    # # duplicate mw_array * vert_count times
    # mw_array = np.tile(mw_array,(vert_count,1)).reshape([-1,16]) # into vert_count*[*16]   # reshape([-1,4,4])   into  4x4 matrix
    # # OBJ_MW_NP = mw_array


    BATCH_FACES = batch_for_shader(shader, 'TRIS', {"pos": VERTICES, "col": face_col, 'nrm': NORMALS}, indices=INDICES,)
    EDGES_BATCH = batch_for_shader(shader, 'LINES', {"pos": VERTICES, "col": edge_col, 'nrm': NORMALS}, indices=mesh.edge_keys)
    VERTS_BATCH = batch_for_shader(shader, 'POINTS', {"pos": VERTICES, "col": vert_col, 'nrm': NORMALS})

    if shared_props.draw_modifiers and len(active_obj.modifiers) > 0:
        obj_eval.to_mesh_clear()
    # else:
    #     mesh.transform(active_obj.matrix_world.inverted())
    PAUSE_HANDLERS = False



def check_obj_updated(depsgraph):
    active_obj = bpy.context.active_object
    if not active_obj:
        return
    global PAUSE_HANDLERS, FORCE_UPDATE_XRAY, TARGET_BVH_LIST, LAST_ACTIVE_OBJ, SINGLE_SKIP_UPDATE
    if PAUSE_HANDLERS:
        return
    if SINGLE_SKIP_UPDATE:
        SINGLE_SKIP_UPDATE = False
        return

    PAUSE_HANDLERS = True


    # Only fire Cache update - when jumping to edit mode.
    if bpy.context.mode != 'EDIT_MESH':
        TARGET_BVH_LIST = None
        # if hasattr(depsgraph, 'updates'):
        #     for update in depsgraph.updates:
        #         # print((f'ID {update.id}, geom {update.is_updated_geometry}, transform: {update.is_updated_transform}, is_eval: {update.id.is_evaluated}'))
        #         if type(update.id) == bpy.types.Object and update.id.type == 'MESH' and (update.is_updated_geometry or update.is_updated_transform):
        #             print('Clear whole cache!') # XXX: rethink if we need auto cache clear...  Maybe just use manual refresh button?
        #             break

    scn_xray_props = bpy.context.scene.xray_props
    obj_xray_props = active_obj.xray_props
    shared_props = scn_xray_props if obj_xray_props.settings_mode == 'GLOBAL' else active_obj.xray_props
    if scn_xray_props.enable_xray and (shared_props.draw_overlay or shared_props.enable_snapping):
        # mark not being in edit mode..
        if active_obj.mode != 'EDIT': # mark going out of edit mode
            global INIT_EDIT_MODE
            INIT_EDIT_MODE = active_obj.mode

        # handle obj mode: selected obj changed > refresh drawn overlay (since no depsgraph updates is fired..)
        if LAST_ACTIVE_OBJ != active_obj.name and active_obj.type == 'MESH' and active_obj.mode in scn_xray_props.show_in_mode:
            LAST_ACTIVE_OBJ = active_obj.name
            refresh_draw_buffers(True)

        if active_obj and active_obj.type == 'MESH':
            if hasattr(depsgraph, 'updates'):
                for update in depsgraph.updates:
                    # print((f'ID {update.id}, geom {update.is_updated_geometry}, transform: {update.is_updated_transform}, is_eval: {update.id.is_evaluated}'))
                    if update.id.name == active_obj.name and active_obj.type == 'MESH':
                        update_pos_cache = update.is_updated_geometry
                        if (update.is_updated_geometry or scn_xray_props.snap_event == 'ALL') and active_obj.mode in ('EDIT'): # or update.is_updated_transform):
                            snap_verts_bm(depsgraph)
                            update_pos_cache = True
                        if active_obj.mode == 'SCULPT':
                            update_pos_cache = True
                        if active_obj.mode in scn_xray_props.show_in_mode:
                            refresh_draw_buffers(update_pos_cache) # out of if - to update on selection change
                        PAUSE_HANDLERS = False
                        return
    PAUSE_HANDLERS = False


# https://computergraphics.stackexchange.com/questions/1718/what-is-the-simplest-way-to-compute-principal-curvature-for-a-mesh-triangle
def vert_convex(v1):
    ''' (n2-n1).dot(p2-p1) / |p2-p1|^2  '''
    # Too much contranst on edges - resutlts in bumpy corners
    curv = 0
    cnt = 0
    v1_snap_co, v1_snap_normal, idx, dist = TARGET_BVH_LIST.find_nearest(v1.co)
    for e in v1.link_edges:
        v2 = e.other_vert(v1)

        v2_snap_co, v2_snap_normal, idx, dist = TARGET_BVH_LIST.find_nearest(v2.co)

        v1_v2 = v2_snap_co-v1_snap_co
        len_sqr = v1_v2.length_squared
        if len_sqr < 0.0001:
            continue
        curv += (v2_snap_normal-v1_snap_normal).dot(v1_v2)/len_sqr
        cnt += 1
    if cnt == 0: # no edges, no curvature..
        return 0
    out = curv/cnt
    return math.sqrt(abs(out)) * (-1 if out < 0 else 1) # squr to try to fade the effect... still not good enough


def vert_cotan_curv(bm):
    # Requires tri mesh, too slow..
    # using cotangent weights
    sel_verts = [v for v in bm.verts if v.select]
    vert_curv = []
    for vi in sel_verts:
        # first calc cell area (simpler barycentric version)
        cell_area = 0
        for f in vi.link_faces:
            cell_area += f.calc_area()
        cell_area /= 3

        # now calc curvature
        dp = Vector((0, 0, 0))
        for e in vi.link_edges:
            vj = e.other_vert(vi)

            # cot a(ij) and cot b(ij)   cot(a) = cos(a)/sin(a) = u.v/|u x v|
            # get link faces, and from there get the other two verts (not vi or vj)
            for face in e.link_faces:
                va = [v for v in face.verts if v != vi and v != e.other_vert(vi)][0]
                va_vi = vi.co - va.co
                va_vj = vj.co - va.co
                cot_a = va_vi.dot(va_vj) / va_vi.cross(va_vj).length

                dp += cot_a * (vj.co - vi.co)

        dp /= 2 * cell_area

        curvature = dp.length / 2 #  |H| = 1/2 |dP|

        # print(f'v[{vi.index}] curvature: {curvature:.3f} [1/mm]')
        vert_curv.append(curvature)
    return vert_curv


def cc_first_iter(bm, sel_verts):
    '''Catmull-Clark 1 iteration - for evaluation of where vert will land after 1st iter of subsurf mod'''

    #Compute the "face points" of all faces connected to selected vertices, indexed by their BMesh indexes.
    face_points = [None] * len(bm.faces)
    for vertex in sel_verts:
        for face in vertex.link_faces:
            if face_points[face.index] is None:
                face_points[face.index] = face.calc_center_median()

    #Compute the "edge midpoints" of all edges connected to the selected vertices, indexed by their BMesh indexes.
    edge_midpoints = [None] * len(bm.edges)
    for vertex in sel_verts:
        for edge in vertex.link_edges:
            if edge_midpoints[edge.index] is None:
                edge_midpoints[edge.index] = 0.5*(edge.verts[0].co + edge.verts[1].co)

    on_surface_iter = []
    #Go through each vertex and compute their smoothed position.
    for vertex in sel_verts:
        if not vertex.is_boundary:
            n = len(vertex.link_faces)
            F = sum((face_points[face.index] for face in vertex.link_faces), Vector()) / n
            R = sum((edge_midpoints[edge.index] for edge in vertex.link_edges), Vector()) / n
            new_co = (F + 2*R + (n - 3)*vertex.co) / n
        else:
            boundary_edges = [e for e in vertex.link_edges if e.is_boundary]  # use list comprehension to filter edges
            boundary_midpoints = [edge_midpoints[e.index] for e in boundary_edges]  # use list comprehension to calculate midpoints
            new_co = 0.5 * vertex.co + 0.25 * sum(boundary_midpoints, Vector())  # use sum function instead of map and Vector() as default argument

        on_surface_iter.append(new_co)

    return on_surface_iter



# cache cc limit factors
a = {n:(n-3)/(n+5) for n in range(1, 10)}
b = {n:4/(n*(n+5)) for n in range(1, 10)}

# https://www.nvidia.com/content/gtc-2010/pdfs/2129_gtc2010.pdf
def cc_limit_aprox(bm, sel_verts):
    '''Catmull-Clark smoothing - for evaluation of where vert will land after subsurf mod'''
    #Compute the "face points" of all faces connected to selected vertices, indexed by their BMesh indexes.
    face_points = [None] * len(bm.faces)
    for vertex in sel_verts:
        for face in vertex.link_faces:
            if face_points[face.index] is None:
                face_points[face.index] = face.calc_center_median()

    #Compute the "edge midpoints" of all edges connected to the selected vertices, indexed by their BMesh indexes.
    edge_midpoints = [None] * len(bm.edges)
    for vertex in sel_verts:
        for edge in vertex.link_edges:
            if edge_midpoints[edge.index] == None:
                edge_midpoints[edge.index] = 0.5*(edge.verts[0].co + edge.verts[1].co)

    on_limit_pos = []
    #Go through each vertex and compute their smoothed position.
    for vertex in sel_verts:
        if not vertex.is_boundary:
            n = len(vertex.link_edges)
            f_cent = sum((face_points[face.index] for face in vertex.link_faces), Vector()) # / n
            e_mid = sum((edge_midpoints[edge.index] for edge in vertex.link_edges), Vector()) # / n
            new_co = a[n] * vertex.co + b[n] * (f_cent + e_mid)

        else: # could not find sollution for boundary verts, using my lucky guess
            # new_co = 0.5 * vertex.co + 0.25 * sum(e_mid, Vector()) # 1st iter - too far
            n = [e for e in vertex.link_edges if e.is_boundary]
            e_mid = [edge_midpoints[e.index] for e in n]
            new_co = 0.4 * vertex.co + 0.3 * sum(e_mid, Vector())


        on_limit_pos.append(new_co)

    return on_limit_pos

# Used if not subsurf mod -  move curved out surface slightly above high poly surface, curved in surface slightly below
# in past u would just offset everything above in uniform way,
def adj_verts_diff_cuvature(TARGET_BVH_LIST, b_vert, b_vert_co, b_vert_norm):
    '''Calc curvature based on difference of adjacent verts co - current vert co'''
    # NOTE:  gives +++ results compared to other curvature methods (cotan, )
    # but results depends on adj.edges count... eg. extruding edge, changes offset - use random sampling - see fun below
    avg = Vector((0, 0, 0))
    for e in b_vert.link_edges:
        other_vert = e.other_vert(b_vert)
        snap_co, normal, idx, dist = TARGET_BVH_LIST.find_nearest(other_vert.co)
        avg += snap_co
    avg = avg/len(b_vert.link_edges) - b_vert_co
    if b_vert_norm != Vector((0,0,0)):
        avg = avg.project(b_vert_norm) # project b_vert norm
        sign = -1 if avg.dot(b_vert_norm) > 0 else 1
        curv = sign*avg.length/2 # h =  ds / 2
        # print(100*curv) # NOTE: should this always be 100
        return 100*curv # what with 100 multiplier?
    else:
        return 1


# XXX: not used...
def adj_verts_diff_cuvature_circle(TARGET_BVH_LIST, b_vert, b_vert_co, b_vert_norm):
    '''Calc curvature based on difference of sampled circle points around b_vert, jumpy results'''
    from statistics import median
    from math import radians
    # NOTE:  gives +++ best results
    avg = Vector((0, 0, 0))
    if len(b_vert.link_edges) > 0:
        edge_lens = [e.calc_length() for e in b_vert.link_edges]
        med_len = median(edge_lens)
    else:
        return 0

    # create 5 samples perpendicular to  b_vert_norm, in radius of med_len
    x_axis = b_vert_norm.cross(Vector((0,0,1)))
    if x_axis.length < 0.0001:
        x_axis = Vector((1,0,0))
    else:
        x_axis.normalize()
    y_axis = b_vert_norm.cross(x_axis)*med_len
    x_axis = x_axis * med_len

    samples = [] # 5 samples on circle
    for i in range(5):
        angle = radians(360/5*i)
        samples.append(b_vert_co + x_axis*math.cos(angle) + y_axis*math.sin(angle))

    global DRAW_POINTS
    DRAW_POINTS = samples
    # for e in b_vert.link_edges:
    #   other_vert = e.other_vert(b_vert)
    #   snap_co, normal, idx, dist = TARGET_BVH_LIST.find_nearest(other_vert.co)
    for sample in samples:
        snap_co, normal, idx, dist = TARGET_BVH_LIST.find_nearest(sample)
        avg += snap_co
    # circle area
    avg = avg/5 - b_vert_co
    if b_vert_norm != Vector((0,0,0)):
        # avg = avg.project(b_vert_norm) # project b_vert norm
        sign = -1 if avg.dot(b_vert_norm) > 0 else 1
        curv = sign*avg.length/2 # h =  ds / 2
        print(100*curv) # NOTE: should this always be 100
        return 100*curv
    else:
        return 1

# bumps get bumped even more.  not good
def shell_factor_curvature(b_vert):
    ''' Use blender vert.calc_shell_factor() - gives resutl from (1,x) range. 1 for flat vert
        Shit - cyclic deps, where bumps get bumped even more...'''
    avg_shell_fac = 0
    if not b_vert.link_edges:
        return 0
    for e in b_vert.link_edges:
        other_vert = e.other_vert(b_vert)
        avg_shell_fac += (other_vert.calc_shell_factor()-1)
    avg_shell_fac = avg_shell_fac/len(b_vert.link_edges) + (b_vert.calc_shell_factor()-1)
    print(avg_shell_fac/2)
    return 100*avg_shell_fac/2 #  (avg + self)/2


def projected_raycast(TARGET_BVH_LIST, ray_origin, target, vert):
    # try  bvhtree ray_cast this time... from camera to vert
    dir = target - ray_origin
    location, normal, face_index, dist = TARGET_BVH_LIST.ray_cast(ray_origin, dir, 1000.0)
    # the if below gives crappy results...
    # if location and normal.normalized().dot(vert.normal) > -0.1: # must hit other side of mesh.. can cause jitter on complex geo..
    #     return None, None, None, None
    return location, normal, face_index, dist


# function that will get adjacent verts for sel_verts
def get_adj_verts(bm,sel_verts):
    adj_verts = []
    # use vert.tag to mark verts that are already in adj_verts
    for v in bm.verts:
        v.tag = False
    for v in sel_verts:
        for e in v.link_edges:
            v_other = e.other_vert(v)
            if not (v_other.select or v_other.tag):
                adj_verts.append(v_other)
                v_other.tag = True
    return adj_verts


def snap_verts_bm(depsgraph):
    context = bpy.context
    scn_xray_props = context.scene.xray_props
    active_obj = context.active_object
    obj_xray_props = active_obj.xray_props

    shared_xray_props = scn_xray_props if obj_xray_props.settings_mode == 'GLOBAL' else obj_xray_props
    if not shared_xray_props.enable_snapping:
        return

    global INIT_EDIT_MODE
    if active_obj.mode != INIT_EDIT_MODE: # do not snap right after entering edit mode.
        INIT_EDIT_MODE = active_obj.mode
        return

    global TARGET_BVH_LIST
    if not TARGET_BVH_LIST:
        TARGET_BVH_LIST = get_obj_mesh_bvht(context, depsgraph, applyModifiers=True, world_space=True)
    if not TARGET_BVH_LIST: # no target was set, or target is not visible..
        print('DrawXray: snap target not found (or is hidden)')
        return

    # global DRAW_POINTS # for debug
    # DRAW_POINTS.clear()

    if shared_xray_props.snap_method == 'PROJECT':
        areas_3d = [] # detect if only one 3d view is present, if so - use it
        init_area = None # else fallback to init_area if available
        for area in context.screen.areas:
            if area.type == 'VIEW_3D': # this will raycans in all 3d views... not what we want, but no way  to detect active
                areas_3d.append(area)
                if area.as_pointer() == AREA_ID: # get area from global var - init at xray enabled.
                    init_area = area

        area = areas_3d[0] if len(areas_3d) == 1 else init_area

        for region in area.regions:
            if region.type == 'WINDOW':
                break
        # view_vector = -1*region.data.view_matrix.inverted().col[2].xyz # z axis
        # two cases - for perspective and ortho
        ray_origin = region.data.view_matrix.inverted().translation
        ray_origin = active_obj.matrix_world.inverted() @ ray_origin

        if not region.data.is_perspective:
            view_dir = region.data.view_matrix.inverted().col[2].xyz
            ray_origin = ray_origin + view_dir * 1000 # move ray origin 1000 units in view direction


    if active_obj.mode == 'EDIT':
        mirror_mod = [mod for mod in active_obj.modifiers if mod.type == 'MIRROR' and mod.mirror_object == None and mod.use_axis[0]]
        subdiv_mod = [mod for mod in active_obj.modifiers if mod.type == 'SUBSURF' and mod.show_viewport]
        subdiv_mod = subdiv_mod[0] if subdiv_mod else None
        offset_norm = 0.1 * shared_xray_props.snap_offset #/ active_obj.scale.length
        # if scn_xray_props.use_simple_snap:
        #     offset_norm *= 0.1

        bm = bmesh.from_edit_mesh(active_obj.data)
        sel_verts = [v for v in bm.verts if v.select]
        # sel_verts_curv = vert_cotan_curv(bm) # not good...

        # grow selection, only good results in snap_method == 'NEAREST'
        if shared_xray_props.snap_method == 'NEAREST' and shared_xray_props.affect_neighbours:
            if len(sel_verts) < len(bm.verts):
                sel_verts.extend(get_adj_verts(bm, sel_verts)) # mark adj verts with tag

        if subdiv_mod: # compensate for subsurf verts movement
            vert_on_cc_surf = cc_limit_aprox(bm, sel_verts) if subdiv_mod.use_limit_surface else cc_first_iter(bm, sel_verts)
            offset_norm *= 0.1 # reducs offset strength when subsurf is on, eye balled value

            for b_vert, vert_cc_limit in zip(sel_verts, vert_on_cc_surf):
                snap_co = None
                if shared_xray_props.snap_method == 'PROJECT':
                    dir = vert_cc_limit - ray_origin
                    snap_co, normal, idx, dist = TARGET_BVH_LIST.ray_cast(ray_origin, dir, 1000.0)
                if not snap_co:  # use 'NEAREST' (also as fallback)
                    snap_co, normal, idx, dist = TARGET_BVH_LIST.find_nearest(vert_cc_limit)

                # if scn_xray_props.use_simple_snap:
                #     b_vert.co += snap_co + normal * offset_norm # scale by active obj.scale ?
                diff = snap_co - vert_cc_limit
                b_vert.co += 1.1*diff + normal * offset_norm # scale by active obj.scale ?

                if mirror_mod and abs(b_vert.co.x) < 2*mirror_mod[0].merge_threshold: #snap to center
                    b_vert.co.x = 0

        else:
            for b_vert in sel_verts:
                snap_co = None
                if shared_xray_props.snap_method == 'PROJECT':
                    dir = b_vert.co - ray_origin
                    snap_co, normal, idx, dist = TARGET_BVH_LIST.ray_cast(ray_origin, dir, 1000.0)
                if not snap_co: # use 'NEAREST'  (also as fallback)
                    snap_co, normal, idx, dist = TARGET_BVH_LIST.find_nearest(b_vert.co)

                # if scn_xray_props.use_simple_snap:
                #     b_vert.co = snap_co + normal * offset_norm # scale by active obj.scale ?
                curv = adj_verts_diff_cuvature(TARGET_BVH_LIST, b_vert, snap_co, normal)
                b_vert.co = snap_co + normal * offset_norm * curv # scale by active obj.scale ?

                if mirror_mod and abs(b_vert.co.x) < 2*mirror_mod[0].merge_threshold: #snap to center
                    b_vert.co.x = 0


        bmesh.update_edit_mesh(active_obj.data)

    elif active_obj.mode == 'SCULPT':
        print("Sculpt mode Xray snapping is not supported") # requires bmesh..  creation may be slow
        pass
        # mod_verts_np = np.empty((len(active_object.data.vertices), 3), 'f')
        # mesh_with_mod.vertices.foreach_get("co", np.reshape(mod_verts_np, len(mesh_with_mod.vertices) * 3))
        # if mirror_mod:
        #     mod = mirror_mod[0]
        #     mod_verts_np[np.abs(mod_verts_np[:, 0]) < 2*mod.merge_threshold, 0] = 0
        # active_object.data.vertices.foreach_set('co', mod_verts_np.ravel())
        # active_object.data.update()

def shrink_transfer_old(active_obj):
    '''Old way - using shrinkwrap modifier'''
    #transfer 'shrink_xray' mod to bm verts
    disabled_mods = []
    # print('Disabling mods')
    for mod in active_obj.modifiers:
        if mod.name != 'shrink_xray' and mod.show_viewport:
            mod.show_viewport = False
            disabled_mods.append(mod)

    # depsgraph = bpy.context.evaluated_depsgraph_get() #to force addon to see disabled mods
    # depsgraph = DEPSGRAPH
    depsgraph.update() #to get disabled mods
    obj_eval = active_obj.evaluated_get(depsgraph)
    active_obj.update_from_editmode()
    mesh_with_mod = obj_eval.to_mesh()

    if active_obj.mode == 'EDIT': #when adding geo, use bm, to get updated vertcount
        bm = bmesh.from_edit_mesh(active_obj.data)
        edit_v_count = len(bm.verts)
    else:
        edit_v_count = len(active_obj.data.vertices)

    if len(mesh_with_mod.vertices) != edit_v_count:
        #print(f'Eval mesh has different vert count. Eval count is: {len(mesh_with_mod.vertices)},  mesh coutn is: {edit_v_count}')
        obj_eval.to_mesh_clear()
        # print('enabling mods')
        for mod in disabled_mods:
            mod.show_viewport = True
        return

    #shrink wrap moves center verts slightly off center. Deal with it below
    mirror_mod = [mod for mod in active_obj.modifiers if mod.type == 'MIRROR' and mod.mirror_object == None and mod.use_axis[0]]

    if active_obj.mode == 'EDIT':
        bm = bmesh.from_edit_mesh(active_obj.data)
        if mirror_mod:
            for b_vert, eval_v in zip(bm.verts, mesh_with_mod.vertices):
                b_vert.co = eval_v.co
                if abs(eval_v.co.x) < 2*mirror_mod[0].merge_threshold: #snap to center
                    b_vert.co.x = 0
        else:
            for b_vert, eval_v in zip(bm.verts, mesh_with_mod.vertices):
                b_vert.co = eval_v.co
        bmesh.update_edit_mesh(active_obj.data)
    elif active_obj.mode == 'SCULPT':
        mod_verts_np = np.empty((len(active_obj.data.vertices), 3), 'f')
        mesh_with_mod.vertices.foreach_get("co", np.reshape(mod_verts_np, len(mesh_with_mod.vertices) * 3))
        if mirror_mod:
            mod = mirror_mod[0]
            mod_verts_np[np.abs(mod_verts_np[:, 0]) < 2*mod.merge_threshold, 0] = 0
        active_obj.data.vertices.foreach_set('co', mod_verts_np.ravel())
        active_obj.data.update()

    obj_eval.to_mesh_clear()
    # print('enabling mods')
    for mod in disabled_mods:
        mod.show_viewport = True


def register():
    bpy.utils.register_class(ScnDrawXrayProps)
    bpy.utils.register_class(XRAY_OT_ObjectPicker)
    bpy.utils.register_class(ObjDrawXrayProps)
    bpy.types.Scene.xray_props = bpy.props.PointerProperty(type=ScnDrawXrayProps)
    bpy.types.Object.xray_props = bpy.props.PointerProperty(type=ObjDrawXrayProps)
    bpy.types.VIEW3D_PT_overlay.append(ShadingXrayPanel)
    bpy.app.handlers.load_post.append(DrawXrayLoadDisable)


def unregister():
    global handle_SpaceView3D
    bpy.types.VIEW3D_PT_overlay.remove(ShadingXrayPanel)
    if handle_SpaceView3D is not None:
        bpy.types.SpaceView3D.draw_handler_remove(handle_SpaceView3D, 'WINDOW')
        handle_SpaceView3D = None

    if xray_scene_update in bpy.app.handlers.depsgraph_update_post:
        bpy.app.handlers.depsgraph_update_post.remove(xray_scene_update)

    bpy.app.handlers.load_post.remove(DrawXrayLoadDisable)

    #line below trigger prop.update=DrawXrayUpdate from xray_props property group
    bpy.context.scene.xray_props.enable_xray = False  #! TO hide draw xray on F8 reload - place alway after removing halders (or crash)
    del bpy.types.Scene.xray_props
    del bpy.types.Object.xray_props
    bpy.utils.unregister_class(ScnDrawXrayProps)
    bpy.utils.unregister_class(XRAY_OT_ObjectPicker)
    bpy.utils.unregister_class(ObjDrawXrayProps)


if __name__ == "__main__":
    register()
